import { ApiProperty, IntersectionType, PartialType } from '@nestjs/swagger';
import { Transform } from 'class-transformer';
import {
  IsArray,
  IsBoolean,
  IsDate,
  IsDateString,
  IsEnum,
  IsIn,
  IsInt,
  IsNumber,
  IsOptional,
  IsString,
  Matches,
  MinLength,
  ValidateIf,
  isNumber
} from 'class-validator';
import dayjs from 'dayjs';
import { PagerDto } from '~/common/dto/pager.dto';
import { MaterialsInOrOutEnum } from '~/constants/enum';
import { Storage } from '~/modules/tools/storage/storage.entity';
import { formatToDate } from '~/utils';

export class MaterialsInOutDto {
  @IsOptional()
  @IsNumber()
  @ApiProperty({ description: '项目Id' })
  projectId?: number;

  @ApiProperty({ description: '产品Id' })
  @ValidateIf(o => !o.inventoryInOutNumber)
  @IsNumber()
  productId: number;

  @ApiProperty({ description: '原材料库存编号' })
  @IsOptional()
  @IsString()
  inventoryInOutNumber: string;

  @ApiProperty({ description: '库存id(产品和单价双主键决定一条库存)' })
  @IsOptional()
  @IsNumber()
  inventoryId: number;

  @ApiProperty({ description: '单位(字典)' })
  @IsNumber()
  @IsOptional()
  unitId: number;

  @ApiProperty({ description: '入库或出库 0:入库 1:出库' })
  @IsEnum(MaterialsInOrOutEnum)
  inOrOut: MaterialsInOrOutEnum;

  @ApiProperty({ description: '时间' })
  @IsOptional()
  time: Date;

  @ApiProperty({ description: '数量' })
  @IsNumber()
  quantity: number;

  @ApiProperty({ description: '单价' })
  @IsOptional()
  @IsNumber()
  unitPrice: number;

  @ApiProperty({ description: '金额' })
  @IsOptional()
  @IsNumber()
  amount: number;

  @ApiProperty({ description: '经办人' })
  @IsOptional()
  @IsString()
  agent: string;

  @ApiProperty({ description: '领料单号' })
  @IsOptional()
  @IsString()
  issuanceNumber: string;

  @ApiProperty({ description: '库存位置' })
  @IsOptional()
  @IsString()
  position: string;

  @IsOptional()
  @IsString()
  @ApiProperty({ description: '备注' })
  remark: string;
}

export class MaterialsInOutUpdateDto extends PartialType(MaterialsInOutDto) {
  @ApiProperty({ description: '附件' })
  @IsOptional()
  @IsArray()
  fileIds: number[];
}
export class MaterialsInOutQueryDto extends PagerDto<MaterialsInOutQueryDto> {
  @ApiProperty({ description: '出入库时间YYYY-MM-DD' })
  @IsOptional()
  // @IsString()
  @Transform(params => {
    // 开始和结束时间用的是一天的开始和一天的结束的时分秒
    return params.value
      ? [
          params.value[0] ? `${formatToDate(params.value[0], 'YYYY-MM-DD')} 00:00:00` : null,
          params.value[1] ? `${formatToDate(params.value[1], 'YYYY-MM-DD')} 23:59:59` : null
        ]
      : [];
  })
  time?: string[];

  @ApiProperty({ description: '入库或出库 0:入库 1:出库' })
  @IsOptional()
  @IsEnum(MaterialsInOrOutEnum)
  inOrOut?: MaterialsInOrOutEnum;

  @ApiProperty({ description: '产品名称' })
  @IsOptional()
  @IsString()
  product?: string;

  @ApiProperty({ description: '经办人' })
  @IsOptional()
  @IsString()
  agent?: string;

  @ApiProperty({ description: '领料单号' })
  @IsOptional()
  @IsString()
  issuanceNumber?: string;

  @ApiProperty({ description: '原材料库存编号' })
  @IsOptional()
  @IsString()
  inventoryInOutNumber?: string;

  @IsOptional()
  @IsString()
  @ApiProperty({ description: '备注' })
  remark?: string;

  @IsOptional()
  @IsNumber()
  @ApiProperty({ description: '项目Id' })
  projectId?: number;

  @IsOptional()
  @IsBoolean()
  @ApiProperty({ description: '是否是用于创建出库记录' })
  isCreateOut?: boolean;
}
