import { Body, Controller, Delete, Get, Param, Post, Put, Query, Res } from '@nestjs/common';
import { ApiOperation, ApiTags } from '@nestjs/swagger';
import { ApiResult } from '~/common/decorators/api-result.decorator';
import { IdParam } from '~/common/decorators/id-param.decorator';
import { Perm, definePermission } from '../auth/decorators/permission.decorator';
import {
  MaterialsInventoryQueryDto,
  MaterialsInventoryDto,
  MaterialsInventoryUpdateDto,
  MaterialsInventoryExportDto
} from '../materials_inventory/materials_inventory.dto';
import { MaterialsInventoryService } from './materials_inventory.service';
import { MaterialsInventoryEntity } from './materials_inventory.entity';
import { ApiSecurityAuth } from '~/common/decorators/swagger.decorator';
import { FastifyReply } from 'fastify';

export const permissions = definePermission('app:materials_inventory', {
  LIST: 'list',
  CREATE: 'create',
  READ: 'read',
  UPDATE: 'update',
  DELETE: 'delete',
  EXPORT: 'export'
} as const);

@ApiTags('MaterialsI Inventory - 原材料库存')
@ApiSecurityAuth()
@Controller('materials-inventory')
export class MaterialsInventoryController {
  constructor(private miService: MaterialsInventoryService) {}

  @Get('export')
  @ApiOperation({ summary: '导出原材料盘点表' })
  @Perm(permissions.EXPORT)
  async exportMaterialsInventoryCheck(
    @Query() dto: MaterialsInventoryExportDto,
    @Res() res: FastifyReply
  ): Promise<void> {
    await this.miService.exportMaterialsInventoryCheck(dto, res);
  }

  @Get()
  @ApiOperation({ summary: '获取原材料库存列表' })
  @ApiResult({ type: [MaterialsInventoryEntity], isPage: true })
  @Perm(permissions.LIST)
  async list(@Query() dto: MaterialsInventoryQueryDto) {
    return this.miService.findAll(dto);
  }

  @Get(':id')
  @ApiOperation({ summary: '获取原材料库存信息' })
  @ApiResult({ type: MaterialsInventoryDto })
  @Perm(permissions.READ)
  async info(@IdParam() id: number) {
    return this.miService.info(id);
  }

  @Post()
  @ApiOperation({ summary: '新增原材料库存' })
  @Perm(permissions.CREATE)
  async create(@Body() dto: MaterialsInventoryDto): Promise<void> {
    await this.miService.create(dto);
  }

  @Put(':id')
  @ApiOperation({ summary: '更新原材料库存' })
  @Perm(permissions.UPDATE)
  async update(@IdParam() id: number, @Body() dto: MaterialsInventoryUpdateDto): Promise<void> {
    await this.miService.update(id, dto);
  }

  @Delete(':id')
  @ApiOperation({ summary: '删除原材料库存' })
  @Perm(permissions.DELETE)
  async delete(@IdParam() id: number): Promise<void> {
    await this.miService.delete(id);
  }
}
