import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { MaterialsInventoryEntity } from './materials_inventory.entity';
import { EntityManager, In, MoreThan, Repository } from 'typeorm';
import {
  MaterialsInventoryDto,
  MaterialsInventoryExportDto,
  MaterialsInventoryQueryDto,
  MaterialsInventoryUpdateDto
} from './materials_inventory.dto';
import { Pagination } from '~/helper/paginate/pagination';
import { FastifyReply } from 'fastify';
import { paginate } from '~/helper/paginate';
import * as ExcelJS from 'exceljs';
import dayjs from 'dayjs';
import { MaterialsInOutEntity } from './in_out/materials_in_out.entity';
import { fieldSearch } from '~/shared/database/field-search';
import { groupBy, sum, uniqBy } from 'lodash';
import { HasInventoryStatusEnum, MaterialsInOrOutEnum, ParamConfigEnum } from '~/constants/enum';
import { ProjectEntity } from '../project/project.entity';
import { calcNumber } from '~/utils';
import { BusinessException } from '~/common/exceptions/biz.exception';
import { ErrorEnum } from '~/constants/error-code.constant';
import { ParamConfigEntity } from '../system/param-config/param-config.entity';
import { isDefined } from 'class-validator';
@Injectable()
export class MaterialsInventoryService {
  constructor(
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(MaterialsInventoryEntity)
    private materialsInventoryRepository: Repository<MaterialsInventoryEntity>,
    @InjectRepository(MaterialsInOutEntity)
    private materialsInOutRepository: Repository<MaterialsInOutEntity>,
    @InjectRepository(ProjectEntity)
    private projectRepository: Repository<ProjectEntity>,
    @InjectRepository(ParamConfigEntity)
    private paramConfigRepository: Repository<ParamConfigEntity>
  ) {}

  /**
   * 导出原材料盘点表
   */
  async exportMaterialsInventoryCheck(
    { time, projectId }: MaterialsInventoryExportDto,
    res: FastifyReply
  ): Promise<void> {
    const ROW_HEIGHT = 20;
    const HEADER_FONT_SIZE = 18;
    const workbook = new ExcelJS.Workbook();
    let projects: ProjectEntity[] = [];
    if (projectId) {
      projects = [await this.projectRepository.findOneBy({ id: projectId })];
    }
    // 查询出项目产品所属的当前库存
    const inventoriesInProjects = await this.materialsInventoryRepository.find({
      where: {
        ...(projects?.length ? { projectId: In(projects.map(item => item.id)) } : null)
      }
    });

    // 生成数据
    const sqb = this.materialsInOutRepository
      .createQueryBuilder('mio')
      .leftJoin('mio.project', 'project')
      .leftJoin('mio.product', 'product')
      .leftJoin('product.unit', 'unit')
      .leftJoin('product.company', 'company')
      .addSelect([
        'project.id',
        'project.name',
        'unit.label',
        'company.name',
        'product.name',
        'product.productSpecification',
        'product.productNumber'
      ])
      .where({
        time: MoreThan(time[0])
      })
      .andWhere('mio.isDelete = 0');

    if (projectId) {
      sqb.andWhere('project.id = :projectId', { projectId });
    }

    const data = await sqb.addOrderBy('mio.time', 'DESC').getMany();
    if (!projectId) {
      projects = uniqBy(
        data.map(item => item.project),
        'id'
      );
    }

    for (const project of projects) {
      const currentProjectInventories = inventoriesInProjects.filter(({ projectId }) =>
        Object.is(projectId, project.id)
      );
      const currentProjectData = data.filter(item => item.projectId === project.id);
      const currentMonthProjectData = currentProjectData.filter(item => {
        return (
          dayjs(item.time).isAfter(dayjs(time[0])) && dayjs(item.time).isBefore(dayjs(time[1]))
        );
      });
      const sheet = workbook.addWorksheet(project.name);
      sheet.mergeCells('A1:T1');
      // 设置标题
      sheet.getCell('A1').value = '山东矿机华信智能科技有限公司原材料盘点表';
      // 设置日期
      sheet.mergeCells('A2:B2');
      sheet.getCell('A2').value = `日期：${dayjs(time[0]).format('YYYY年M月')}`;
      // 设置表头
      const headers = [
        '出入库单号',
        '公司名称',
        '产品名称',
        '单位',
        '库存数量',
        '单价',
        '金额',
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '结存数量',
        '单价',
        '金额',
        '经办人',
        '领料单号',
        '备注'
      ];
      sheet.addRow(headers);
      sheet.addRow([
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '入库时间',
        '数量',
        '单价',
        '金额',
        '出库时间',
        '数量',
        '单价',
        '金额',
        '',
        '',
        '',
        '',
        '',
        ''
      ]);
      for (let i = 1; i <= 7; i++) {
        sheet.mergeCells(`${String.fromCharCode(64 + i)}3:${String.fromCharCode(64 + i)}4`);
      }
      // 入库
      sheet.mergeCells('H3:K3');
      sheet.getCell('H3').value = '入库';

      // 出库
      sheet.mergeCells('L3:O3');
      sheet.getCell('L3').value = '出库';

      for (let i = 8; i <= 15; i++) {
        sheet.getCell(`${String.fromCharCode(64 + i)}4`).style.fill = {
          type: 'pattern',
          pattern: 'solid',
          fgColor: { argb: 'FFFFC000' }
        };
      }

      for (let i = 16; i <= 21; i++) {
        sheet.mergeCells(`${String.fromCharCode(64 + i)}3:${String.fromCharCode(64 + i)}4`);
      }

      // 固定信息样式设定
      sheet.eachRow((row, index) => {
        row.alignment = { vertical: 'middle', horizontal: 'center' };
        row.font = { bold: true };
        row.height = ROW_HEIGHT;
        if (index >= 3) {
          row.eachCell(cell => {
            cell.border = {
              top: { style: 'thin' },
              left: { style: 'thin' },
              bottom: { style: 'thin' },
              right: { style: 'thin' }
            };
          });
        }
      });

      const groupedData = groupBy<MaterialsInOutEntity>(
        currentMonthProjectData,
        record => `${record.productId}-${record.unitPrice}`
      );
      let number = 0;
      const groupedInventories = groupBy(
        currentProjectInventories,
        item => `${item.projectId}_${item.productId}`
      );
      for (const key in groupedData) {
        // 目前暂定逻辑出库只有一次或者没有出库。不会对一个入库的记录多次出库,故而用find。
        const inRecord = groupedData[key].find(item => item.inOrOut === MaterialsInOrOutEnum.In);
        const outRecord = groupedData[key].find(item => item.inOrOut === MaterialsInOrOutEnum.Out);
        const currInventories =
          groupedInventories[`${inRecord.projectId}_${inRecord.productId}`]?.shift();
        const allDataFromMonth = data.filter(
          res => res.projectId === inRecord.projectId && res.productId === inRecord.productId
        );
        let currentQuantity = 0;
        let balanceQuantity = 0;
        // 月初库存数量
        if (currInventories) {
          const sumIn = sum(
            allDataFromMonth
              .filter(res => Object.is(res.inOrOut, MaterialsInOrOutEnum.In))
              .map(item => item.quantity)
          );
          const sumOut = sum(
            allDataFromMonth
              .filter(res => Object.is(res.inOrOut, MaterialsInOrOutEnum.Out))
              .map(item => item.quantity)
          );
          const sumDistance = calcNumber(sumIn, sumOut, 'subtract');
          currentQuantity = calcNumber(currInventories.quantity, sumDistance, 'subtract');
        }
        // 结存库存数量
        balanceQuantity = calcNumber(
          currentQuantity,
          calcNumber(inRecord.quantity, outRecord?.quantity || 0, 'subtract'),
          'add'
        );
        number++;
        sheet.addRow([
          `${inRecord.inventoryInOutNumber || ''}`,
          inRecord.product.company.name || '',
          inRecord.product.name || '',
          inRecord.product.unit.label || '',
          currentQuantity,
          parseFloat(`${inRecord.unitPrice || 0}`),
          calcNumber(currentQuantity, inRecord.unitPrice || 0, 'multiply'),
          inRecord.time,
          inRecord.quantity || 0,
          parseFloat(`${inRecord.unitPrice || 0}`),
          parseFloat(`${inRecord.amount || 0}`),
          outRecord?.time || '',
          outRecord?.quantity || 0,
          parseFloat(`${outRecord?.unitPrice || 0}`),
          parseFloat(`${outRecord?.amount || 0}`),
          balanceQuantity,
          parseFloat(`${inRecord?.unitPrice || 0}`),
          calcNumber(balanceQuantity, inRecord?.unitPrice || 0, 'multiply'),
          `${inRecord?.agent || ''}/${outRecord?.agent || ''}`,
          outRecord?.issuanceNumber || '',
          `${inRecord?.remark || ''}/${outRecord?.remark || ''}`
        ]);
      }
      sheet.getCell('A1').font = { size: HEADER_FONT_SIZE };

      // 固定信息样式设定
      sheet.eachRow((row, index) => {
        if (index >= 5) {
          row.alignment = { vertical: 'middle', horizontal: 'center' };
          row.height = ROW_HEIGHT;
          row.eachCell(cell => {
            cell.border = {
              top: { style: 'thin' },
              left: { style: 'thin' },
              bottom: { style: 'thin' },
              right: { style: 'thin' }
            };
          });
        }
      });

      sheet.columns.forEach((column, index: number) => {
        let maxColumnLength = 0;
        const autoWidth = ['B', 'C', 'S', 'U'];
        if (String.fromCharCode(65 + index) === 'B') maxColumnLength = 20;
        if (autoWidth.includes(String.fromCharCode(65 + index))) {
          column.eachCell({ includeEmpty: true }, (cell, rowIndex) => {
            if (rowIndex >= 5) {
              const columnLength = `${cell.value || ''}`.length;
              if (columnLength > maxColumnLength) {
                maxColumnLength = columnLength;
              }
            }
          });
          column.width = maxColumnLength < 12 ? 12 : maxColumnLength; // Minimum width of 10
        } else {
          column.width = 12;
        }
      });
    }
    //读取buffer进行传输
    const buffer = await workbook.xlsx.writeBuffer();
    res
      .header('Content-Type', 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
      .header(
        'Content-Disposition',
        `attachment; filename="${encodeURIComponent('导出_excel' + new Date().getTime() + '.xls')}"`
      )
      .send(buffer);
  }

  /**
   * 查询所有盘点信息
   */
  async findAll({
    page,
    pageSize,
    product,
    keyword,
    isHasInventory
  }: MaterialsInventoryQueryDto): Promise<Pagination<MaterialsInventoryEntity>> {
    const queryBuilder = this.materialsInventoryRepository
      .createQueryBuilder('materialsInventory')
      .leftJoin('materialsInventory.project', 'project')
      .leftJoin('materialsInventory.product', 'product')
      .leftJoin('product.unit', 'unit')
      .leftJoin('product.company', 'company')
      .addSelect([
        'project.name',
        'project.id',
        'unit.id',
        'unit.label',
        'company.id',
        'company.name',
        'product.id',
        'product.name',
        'product.productSpecification',
        'product.productNumber'
      ])
      .where('materialsInventory.isDelete = 0');
    if (product) {
      queryBuilder.andWhere('product.name like :product', { product: `%${product}%` });
    }
    if (keyword) {
      queryBuilder.andWhere(
        '(materialsInventory.inventoryNumber like :keyword or  product.name like :keyword or product.productNumber like :keyword or product.productSpecification like :keyword)',
        {
          keyword: `%${keyword}%`
        }
      );
    }
    if (isHasInventory == HasInventoryStatusEnum.Yes) {
      queryBuilder.andWhere('materialsInventory.quantity > 0');
    }
    if (isHasInventory == HasInventoryStatusEnum.No) {
      queryBuilder.andWhere('materialsInventory.quantity = 0');
    }
    return paginate<MaterialsInventoryEntity>(queryBuilder, {
      page,
      pageSize
    });
  }

  /**
   * 新增库存
   */
  async create(dto: MaterialsInventoryDto): Promise<void> {
    await this.materialsInventoryRepository.insert(dto);
  }

  /**
   * 更新库存
   */
  async update(id: number, data: Partial<MaterialsInventoryUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      await manager.update(MaterialsInventoryEntity, id, {
        ...data
      });
    });
  }

  /**
   * 产品入库后计算最新库存
   * 请注意。产品库存需要根据产品id和价格双主键存储。因为产品价格会变化，需要分开统计。
   * @param data 传入项目ID,产品ID和入库数量和单价
   * @param manager 传入事务对象防止开启多重事务
   */
  async inInventory(
    data: {
      position?: string;
      projectId: number;
      productId: number;
      quantity: number;
      unitPrice?: number;
      changedUnitPrice?: number;
    },
    manager: EntityManager
  ): Promise<MaterialsInventoryEntity> {
    const {
      projectId,
      productId,
      quantity: inQuantity,
      unitPrice,
      changedUnitPrice,
      position
    } = data;

    const exsitedInventory = await manager.findOne(MaterialsInventoryEntity, {
      where: { projectId, productId, unitPrice }, // 根据项目，产品，价格查出之前的实时库存情况
      lock: { mode: 'pessimistic_write' } // 开启悲观行锁，防止脏读和修改
    });

    // 若不存在库存，直接新增库存
    if (!exsitedInventory) {
      const inventoryNumber = await this.generateInventoryNumber();
      const { raw } = await manager.insert(MaterialsInventoryEntity, {
        projectId,
        productId,
        unitPrice,
        inventoryNumber,
        position,
        quantity: inQuantity
      });
      return manager.findOne(MaterialsInventoryEntity, { where: { id: raw.insertId } });
    }
    // 若该项目存在库存，则该项目该产品的库存增加
    let { quantity, id } = exsitedInventory;
    const newQuantity = calcNumber(quantity || 0, inQuantity || 0, 'add');
    if (isNaN(newQuantity)) {
      throw new Error('库存数量不合法');
    }
    await manager.update(MaterialsInventoryEntity, id, {
      quantity: newQuantity,
      unitPrice: changedUnitPrice || undefined
    });

    return manager.findOne(MaterialsInventoryEntity, { where: { id } });
  }

  /**
   * 产品出库
   * @param data 传入库存ID(一定存在。)
   * @param manager 传入事务对象防止开启多重事务
   */
  async outInventory(
    data: {
      quantity: number;
      inventoryId?: number;
      productId: number;
      unitPrice?: number;
    },
    manager: EntityManager
  ): Promise<MaterialsInventoryEntity> {
    const { quantity: outQuantity, inventoryId, productId, unitPrice } = data;
    let searchPayload: any = {};
    if (inventoryId) {
      searchPayload.id = inventoryId;
    } else {
      // 删除出入库记录时，需要根据产品ID和价格查找库存
      searchPayload = { productId, unitPrice };
    }
    // 开启悲观行锁，防止脏读和修改
    const inventory = await manager.findOne(MaterialsInventoryEntity, {
      where: searchPayload,
      lock: { mode: 'pessimistic_write' }
    });
    // 检查库存剩余
    if (inventory.quantity < outQuantity) {
      throw new BusinessException(ErrorEnum.INVENTORY_INSUFFICIENT);
    }
    // 若该项目的该产品库存充足，则该项目该产品的库存减少
    let { quantity, id } = inventory;
    const newQuantity = calcNumber(quantity || 0, outQuantity || 0, 'subtract');
    if (isNaN(newQuantity)) {
      throw new Error('库存数量不足。请检查库存或重新操作。');
    }
    await manager.update(MaterialsInventoryEntity, id, {
      quantity: newQuantity
    });

    return manager.findOne(MaterialsInventoryEntity, { where: { id } });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    // 比较重要，做逻辑删除
    await this.materialsInventoryRepository.update(id, { isDelete: 1 });
  }

  /**
   * 获取某个价格的某个商品库存信息
   */
  async info(id: number) {
    const info = await this.materialsInventoryRepository
      .createQueryBuilder('materialsInventory')
      .leftJoin('materialsInventory.project', 'project')
      .leftJoin('materialsInventory.product', 'product')
      .leftJoin('product.unit', 'unit')
      .leftJoin('product.company', 'company')
      .addSelect([
        'project.name',
        'project.id',
        'product.id',
        'product.name',
        'unit.label',
        'company.name',
        'product.productSpecification',
        'product.productNumber'
      ])
      .where({
        id
      })
      .andWhere('materialsInventory.isDelete = 0')
      .getOne();
    return info;
  }

  /**
   * 生成库存编号
   * @returns 库存编号
   */
  async generateInventoryNumber() {
    const prefix =
      (
        await this.paramConfigRepository.findOne({
          where: {
            key: ParamConfigEnum.InventoryNumberPrefix
          }
        })
      )?.value || '';
    const lastInventory = await this.materialsInventoryRepository
      .createQueryBuilder('materials_inventory')
      .select(
        `MAX(CAST(REPLACE(materials_inventory.inventoryNumber, '${prefix}', '') AS UNSIGNED))`,
        'maxInventoryNumber'
      )
      .getRawOne();
    const lastNumber = lastInventory.maxInventoryNumber
      ? parseInt(lastInventory.maxInventoryNumber.replace(prefix, ''))
      : 0;
    const newNumber = lastNumber + 1 < 1000 ? 1000 : lastNumber + 1;
    return `${prefix}${newNumber}`;
  }
}
