import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { ProductEntity } from './product.entity';
import { EntityManager, Like, Repository } from 'typeorm';
import { ProductDto, ProductQueryDto, ProductUpdateDto } from './product.dto';
import { Pagination } from '~/helper/paginate/pagination';
import { paginate } from '~/helper/paginate';
import { Storage } from '../tools/storage/storage.entity';
import { BusinessException } from '~/common/exceptions/biz.exception';
import { ErrorEnum } from '~/constants/error-code.constant';
import { fieldSearch } from '~/shared/database/field-search';
import { ParamConfigEnum } from '~/constants/enum';
import { ParamConfigEntity } from '../system/param-config/param-config.entity';

@Injectable()
export class ProductService {
  constructor(
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(ProductEntity)
    private productRepository: Repository<ProductEntity>,
    @InjectRepository(Storage)
    private storageRepository: Repository<Storage>,
    @InjectRepository(ParamConfigEntity)
    private paramConfigRepository: Repository<ParamConfigEntity>
  ) {}

  /**
   * 查询所有产品
   */
  async findAll({
    page,
    pageSize,
    ...fields
  }: ProductQueryDto): Promise<Pagination<ProductEntity>> {
    const { company: companyName, keyword, ...ext } = fields;
    const sqb = this.productRepository
      .createQueryBuilder('product')
      .leftJoin('product.files', 'files')
      .leftJoin('product.company', 'company')
      .leftJoin('product.unit', 'unit')
      .addSelect(['files.id', 'files.path', 'company.name', 'company.id', 'unit.id', 'unit.label'])
      .where(fieldSearch(ext))
      .andWhere('product.isDelete = 0')
      .addOrderBy('product.namePinyin', 'ASC');
    if (companyName) {
      sqb.andWhere({
        company: {
          name: Like(`%${companyName}%`)
        }
      });
    }
    if (keyword) {
      //关键字模糊查询product的name,productNumber,productSpecification
      sqb.andWhere(
        '(product.name like :keyword or product.productNumber like :keyword or product.productSpecification like :keyword)',
        {
          keyword: `%${keyword}%`
        }
      );
   
    }
    return paginate<ProductEntity>(sqb, {
      page,
      pageSize
    });
  }

  /**
   * 新增
   */
  async create(dto: ProductDto): Promise<void> {
    const { name, companyId } = dto;
    const isExsit = await this.productRepository.findOne({
      where: { name, company: { id: companyId } }
    });
    if (isExsit) {
      throw new BusinessException(ErrorEnum.PRODUCT_EXIST);
    }
    await this.productRepository.insert(
      this.productRepository.create({ ...dto, productNumber: await this.generateProductNumber() })
    );
  }

  /**
   * 更新
   */
  async update(id: number, { fileIds, ...data }: Partial<ProductUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      await manager.update(
        ProductEntity,
        id,
        this.productRepository.create({
          ...data
        })
      );
      const product = await this.productRepository
        .createQueryBuilder('product')
        .leftJoinAndSelect('product.files', 'files')
        .where('product.id = :id', { id })
        .getOne();
      if (fileIds?.length) {
        const count = await this.storageRepository
          .createQueryBuilder('storage')
          .where('storage.id in(:fileIds)', { fileIds })
          .getCount();
        if (count !== fileIds?.length) {
          throw new BusinessException(ErrorEnum.STORAGE_NOT_FOUND);
        }
        // 附件要批量插入
        await manager.createQueryBuilder().relation(ProductEntity, 'files').of(id).add(fileIds);
      }
    });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    await this.productRepository.update(id, { isDelete: 1 });
  }

  /**
   * 获取单个产品信息
   */
  async info(id: number) {
    const info = await this.productRepository
      .createQueryBuilder('product')
      .leftJoin('product.company', 'company')
      .addSelect(['company.name', 'company.id'])
      .where({
        id
      })
      .andWhere('product.isDelete = 0')
      .getOne();
    return info;
  }

  /**
   * 解除附件关联
   * @param id 产品ID
   * @param fileIds 附件ID
   */
  async unlinkAttachments(id: number, fileIds: number[]) {
    await this.entityManager.transaction(async manager => {
      const product = await this.productRepository
        .createQueryBuilder('product')
        .leftJoinAndSelect('product.files', 'files')
        .where('product.id = :id', { id })
        .getOne();
      const linkedFiles = product.files
        .map(item => item.id)
        .filter(item => !fileIds.includes(item));
      // 附件要批量更新
      await manager
        .createQueryBuilder()
        .relation(ProductEntity, 'files')
        .of(id)
        .addAndRemove(linkedFiles, product.files);
    });
  }

  /**
   * 生成产品编号
   * @returns 产品编号
   */
  async generateProductNumber(): Promise<string> {
    const prefix =
      (
        await this.paramConfigRepository.findOne({
          where: {
            key: ParamConfigEnum.ProductNumberPrefix
          }
        })
      )?.value || '';
    const lastProduct = await this.productRepository
      .createQueryBuilder('product')
      .select(
        `MAX(CAST(REPLACE(COALESCE(product.product_number, ''), '${prefix}', '') AS UNSIGNED))`,
        'productNumber'
      )
      .getRawOne();
    const lastNumber = lastProduct.productNumber
      ? parseInt(lastProduct.productNumber.replace(prefix, ''))
      : 0;
    const newNumber = lastNumber + 1 < 1000 ? 1000 : lastNumber + 1;
    return `${prefix}${newNumber}`;
  }
}
