import {
  Body,
  Controller,
  Get,
  Query,
  Put,
  Delete,
  Post,
  BadRequestException
} from '@nestjs/common';
import { ApiOperation, ApiTags } from '@nestjs/swagger';
import { Perm, definePermission } from '../auth/decorators/permission.decorator';
import { ApiSecurityAuth } from '~/common/decorators/swagger.decorator';
import { ProjectService } from './project.service';
import { ApiResult } from '~/common/decorators/api-result.decorator';
import { ProjectEntity } from './project.entity';
import { ProjectDto, ProjectQueryDto, ProjectUpdateDto } from './project.dto';
import { IdParam } from '~/common/decorators/id-param.decorator';
export const permissions = definePermission('app:project', {
  LIST: 'list',
  CREATE: 'create',
  READ: 'read',
  UPDATE: 'update',
  DELETE: 'delete'
} as const);

@ApiTags('Project - 项目')
@ApiSecurityAuth()
@Controller('project')
export class ProjectController {
  constructor(private projectService: ProjectService) {}

  @Get()
  @ApiOperation({ summary: '分页获取项目列表' })
  @ApiResult({ type: [ProjectEntity], isPage: true })
  @Perm(permissions.LIST)
  async list(@Query() dto: ProjectQueryDto) {
    return this.projectService.findAll(dto);
  }

  @Get(':id')
  @ApiOperation({ summary: '获取项目信息' })
  @ApiResult({ type: ProjectDto })
  @Perm(permissions.READ)
  async info(@IdParam() id: number) {
    return this.projectService.info(id);
  }

  @Post()
  @ApiOperation({ summary: '新增项目' })
  @Perm(permissions.CREATE)
  async create(@Body() dto: ProjectDto): Promise<void> {
    await this.projectService.create(dto);
  }

  @Put(':id')
  @ApiOperation({ summary: '更新项目' })
  @Perm(permissions.UPDATE)
  async update(@IdParam() id: number, @Body() dto: ProjectUpdateDto): Promise<void> {
    await this.projectService.update(id, dto);
  }

  @Delete(':id')
  @ApiOperation({ summary: '删除项目' })
  @Perm(permissions.DELETE)
  async delete(@IdParam() id: number): Promise<void> {
    await this.projectService.delete(id);
  }

  @Put('unlink-attachments/:id')
  @ApiOperation({ summary: '附件解除关联' })
  @Perm(permissions.UPDATE)
  async unlinkAttachments(
    @IdParam() id: number,
    @Body() { fileIds }: ProjectUpdateDto
  ): Promise<void> {
    await this.projectService.unlinkAttachments(id, fileIds);
  }
}
