import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { ProjectEntity } from './project.entity';
import { EntityManager, Like, Repository } from 'typeorm';
import { ProjectDto, ProjectQueryDto, ProjectUpdateDto } from './project.dto';
import { Pagination } from '~/helper/paginate/pagination';
import { paginate } from '~/helper/paginate';
import { Storage } from '../tools/storage/storage.entity';
import { BusinessException } from '~/common/exceptions/biz.exception';
import { ErrorEnum } from '~/constants/error-code.constant';
import { fieldSearch } from '~/shared/database/field-search';

@Injectable()
export class ProjectService {
  constructor(
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(ProjectEntity)
    private projectRepository: Repository<ProjectEntity>,
    @InjectRepository(Storage)
    private storageRepository: Repository<Storage>
  ) {}

  /**
   * 分页查询所有
   */
  async findAll({
    page,
    pageSize,
    ...fields
  }: ProjectQueryDto): Promise<Pagination<ProjectEntity>> {
    const queryBuilder = this.projectRepository
      .createQueryBuilder('project')
      .leftJoin('project.files', 'files')
      .addSelect(['files.id', 'files.path'])
      .where(fieldSearch(fields))
      .andWhere('project.isDelete = 0');

    return paginate<ProjectEntity>(queryBuilder, {
      page,
      pageSize
    });
  }

  /**
   * 新增
   */
  async create(dto: ProjectDto): Promise<void> {
    await this.projectRepository.insert(dto);
  }

  /**
   * 更新
   */
  async update(id: number, { fileIds, ...data }: Partial<ProjectUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      await manager.update(ProjectEntity, id, {
        ...data
      });
      const project = await this.projectRepository
        .createQueryBuilder('project')
        .leftJoinAndSelect('project.files', 'files')
        .where('project.id = :id', { id })
        .getOne();
      if (fileIds?.length) {
        const count = await this.storageRepository
          .createQueryBuilder('storage')
          .where('storage.id in(:fileIds)', { fileIds })
          .getCount();
        if (count !== fileIds?.length) {
          throw new BusinessException(ErrorEnum.STORAGE_NOT_FOUND);
        }
        // 附件要批量插入
        await manager
          .createQueryBuilder()
          .relation(ProjectEntity, 'files')
          .of(id)
          .add(fileIds);
      }
    });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    // 比较重要，做逻辑删除
    await this.projectRepository.update(id, { isDelete: 1 });
  }

  /**
   * 获取单个信息
   */
  async info(id: number) {
    const info = await this.projectRepository
      .createQueryBuilder('project')
      .where({
        id
      })
      .andWhere('project.isDelete = 0')
      .getOne();
    return info;
  }

  /**
   * 解除附件关联
   * @param id 实体ID
   * @param fileIds 附件ID
   */
  async unlinkAttachments(id: number, fileIds: number[]) {
    await this.entityManager.transaction(async manager => {
      const project = await this.projectRepository
        .createQueryBuilder('project')
        .leftJoinAndSelect('project.files', 'files')
        .where('project.id = :id', { id })
        .getOne();
      const linkedFiles = project.files
        .map(item => item.id)
        .filter(item => !fileIds.includes(item));
      // 附件要批量更新
      await manager
        .createQueryBuilder()
        .relation(ProjectEntity, 'files')
        .of(id)
        .addAndRemove(linkedFiles, project.files);
    });
  }
}
