import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { EntityManager, Repository, SelectQueryBuilder } from 'typeorm';
import { VehicleUsageEntity } from './vehicle_usage.entity';
import { Storage } from '../tools/storage/storage.entity';
import { ErrorEnum } from '~/constants/error-code.constant';
import { BusinessException } from '~/common/exceptions/biz.exception';
import { paginate } from '~/helper/paginate';
import { Pagination } from '~/helper/paginate/pagination';
import { fieldSearch } from '~/shared/database/field-search';
import { VehicleUsageQueryDto, VehicleUsageDto } from './vehicle_usage.dto';
import { VehicleUsageUpdateDto } from './vehicle_usage.dto';

@Injectable()
export class VehicleUsageService {
  constructor(
    @InjectRepository(VehicleUsageEntity)
    private vehicleUsageRepository: Repository<VehicleUsageEntity>,
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(Storage)
    private storageRepository: Repository<Storage>
  ) {}

  /**
   * 分页查询所有
   */
  async findAll({
    page,
    pageSize,

    ...fields
  }: VehicleUsageQueryDto): Promise<Pagination<VehicleUsageEntity>> {
    const { vehicle, ...ext } = fields;
    const sqb = this.buildLeftJoinRelations().where(fieldSearch(ext));
    if (vehicle) {
      sqb.andWhere('vehicle.label like :vehicleName', { vehicleName: `%${vehicle}%` });
    }
    return paginate<VehicleUsageEntity>(sqb, {
      page,
      pageSize
    });
  }

  /**
   * 新增
   */
  async create(dto: VehicleUsageDto): Promise<void> {
    // const { name, companyId } = dto;
    // const isExsit = await this.vehicleUsageRepository.findOne({
    //   where: { name, company: { id: companyId } }
    // });
    // if (isExsit) {
    //   throw new BusinessException(ErrorEnum.PRODUCT_EXIST);
    // }
    await this.vehicleUsageRepository.insert(this.vehicleUsageRepository.create(dto));
  }

  /**
   * 更新
   */
  async update(id: number, { fileIds, ...data }: Partial<VehicleUsageUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      await manager.update(
        VehicleUsageEntity,
        id,
        this.vehicleUsageRepository.create({
          ...data
        })
      );
      const vehicleUsage = await this.vehicleUsageRepository
        .createQueryBuilder('vehicle_usage')
        .leftJoinAndSelect('vehicle_usage.files', 'files')
        .where('vehicle_usage.id = :id', { id })
        .getOne();
      if (fileIds?.length) {
        const count = await this.storageRepository
          .createQueryBuilder('storage')
          .where('storage.id in(:fileIds)', { fileIds })
          .getCount();
        if (count !== fileIds?.length) {
          throw new BusinessException(ErrorEnum.STORAGE_NOT_FOUND);
        }
        // 附件要批量插入
        await manager
          .createQueryBuilder()
          .relation(VehicleUsageEntity, 'files')
          .of(id)
          .add(fileIds);
      }
    });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    await this.vehicleUsageRepository.delete(id);
  }

  /**
   * 获取单个信息
   */
  async info(id: number) {
    const info = this.buildLeftJoinRelations()
      .where({
        id
      })
      .getOne();
    return info;
  }

  /**
   * 解除附件关联
   * @param id 记录ID
   * @param fileIds 附件ID
   */
  async unlinkAttachments(id: number, fileIds: number[]) {
    await this.entityManager.transaction(async manager => {
      const vehicle_usage = await this.vehicleUsageRepository
        .createQueryBuilder('vehicle_usage')
        .leftJoinAndSelect('vehicle_usage.files', 'files')
        .where('vehicle_usage.id = :id', { id })
        .getOne();
      const linkedFiles = vehicle_usage.files
        .map(item => item.id)
        .filter(item => !fileIds.includes(item));
      // 附件要批量更新
      await manager
        .createQueryBuilder()
        .relation(VehicleUsageEntity, 'files')
        .of(id)
        .addAndRemove(linkedFiles, vehicle_usage.files);
    });
  }

  /**
   * 封装和查询关联关系
   */
  buildLeftJoinRelations() {
    return this.vehicleUsageRepository
      .createQueryBuilder('vehicle_usage')
      .leftJoin('vehicle_usage.files', 'files')
      .leftJoin('vehicle_usage.vehicle', 'vehicle')
      .addSelect(['files.id', 'files.path', 'vehicle.id', 'vehicle.label']);
  }
}
